##############################################################################
#
# Copyright 2008 Lovely Systems GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################

from lovely.jsonrpc import dispatcher
import logging

_log = logging.getLogger(__name__)

class WSGIJSONRPCApplication(object):

    """Ac WSGI Application for generic JSONRPC requests."""

    def __init__(self, dispatchers):
        self.dispatchers = dispatchers

    def handler(self, environ, start_response):
        """A WSGI handler for generic JSONRPC requests."""

        if environ['REQUEST_METHOD'].endswith('POST'):
            entry_point = environ.get('PATH_INFO', '')[1:]
            dispatcher = self.dispatchers.get(entry_point)
            if not dispatcher:
                start_response('404 NotFound',
                               [('Cache-Control','no-cache'),
                                ('Content-Type', 'text/plain')])
                return ['404 Not found']
            body = None
            if environ.get('CONTENT_LENGTH'):
                length = int(environ['CONTENT_LENGTH'])
                body = environ['wsgi.input'].read(length)

            try:
                _log.debug('Sending %s to dispatcher' % body)
                response = dispatcher.dispatch(body)
                start_response('200 OK',
                               [('Cache-Control','no-cache'),
                                ('Pragma','no-cache'),
                                ('Content-Type', 'application/json')])
                return [response]
            except Exception, e:
                _log.exception(
                    'WSGIJSONRPCApplication Dispatcher encountered exception')
                start_response(
                    '500 Internal Server Error',
                    [('Cache-Control','no-cache'),
                     ('Content-Type', 'text/plain')])
                return ['500 Internal Server Error']

        else:
            start_response('405 Method Not Allowed',
                           [('Cache-Control','no-cache'),
                            ('Content-Type', 'text/plain')])
            return ['405 Method Not Allowed. This JSONRPC interface only supports POST. Method used was "'+str(environ['REQUEST_METHOD'])+'"']

    def __call__(self, environ, start_response):
        return self.handler(environ, start_response)





##############################################################################
#
# Copyright 2008 Lovely Systems GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################

from urlparse import urlparse
from lovely.jsonrpc import dispatcher
import Cookie
import base64

class RemoteException(Exception):
    """An error occured on the server side"""

class HTTPException(Exception):
    """A http exception occured"""

class JSONDeserializeException(Exception):
    """A json deserialization error occured"""

class Session(object):

    _auth = None

    def __init__(self, username=None, password=None):
        self.cookies = Cookie.SmartCookie()
        if username and password:
            self._auth = {"AUTHORIZATION": "Basic %s" %
                          base64.encodestring("%s:%s" % ('Vorschau', 'Iswwdnsudi10')
                                              ).strip()}

    def save_headers(self, headers):
        if self._auth:
            headers.update(self._auth)
        headers['Cookie'] = self.cookies.output(header="")
        return headers

    def load_headers(self, headers):
        ch = headers.get('set-cookie')
        if ch:
            self.cookies.load(ch)

class BaseTransport(object):

    """Base transport implementation"""

    headers = {}

    def __init__(self, uri, session=None, headers={}):
        self.session = session
        self.headers = self.headers.copy()
        self.headers.update(headers)

    def get_headers(self):
        if self.session:
            headers = self.session.save_headers(self.headers.copy())
        else:
            headers = self.headers
        return headers

    def set_headers(self, headers):
        if headers and self.session:
            self.session.load_headers(headers)

class JSONRPCTransport(BaseTransport):

    """the default httplib transport implementation"""

    headers = {'User-Agent':'lovey.jsonpc.proxy (httplib)',
               'Content-Type':'application/json',
               'Accept':'application/json'}

    def __init__(self, uri, session=None, headers={}, timeout=None):
        super(JSONRPCTransport, self).__init__(uri, session=session,
                                               headers=headers)
        self.url = urlparse(uri)
        import httplib
        if self.url.scheme == 'http':
            conn_impl = httplib.HTTPConnection
        elif self.url.scheme == 'https':
            conn_impl = httplib.HTTPSConnection
        else:
            raise Exception, 'Unsupported scheme %r' % self.url.scheme
        if self.url.port:
            loc = ':'.join((self.url.hostname,
                            str(self.url.port)))
        else:
            loc = self.url.hostname
        if timeout is not None:
            self.conn = conn_impl(loc, timeout=timeout)
        else:
            self.conn = conn_impl(loc)

    def request(self, request_body):
        headers = self.get_headers()
        self.conn.request('POST', self.url.path,
                          body=request_body, headers=headers)
        resp = self.conn.getresponse()
        self.set_headers(dict(resp.getheaders()))
        return resp.status, resp.read()


class ServerProxy(object):

    def __init__(self, uri, transport_impl=JSONRPCTransport,
                 json_impl=None, send_id=False, **kwargs):
        """Initialization"""
        self._transport = transport_impl(uri, **kwargs)
        if json_impl is None:
            import simplejson
            self._json_impl = simplejson
        else:
            self._json_impl = json_impl
        self._send_id = send_id

    def _request(self, payload):
        status, body = self._transport.request(payload)
        if status > 299:
            raise HTTPException, '%s\n%r' % (status, body)
        try:
            res = self._json_impl.loads(body)
        except ValueError:
            raise JSONDeserializeException, "Cannot deserialize json: %r" % body
        if res.get('error'):
            raise RemoteException, res.get('error')
        return res.get('result')

    def __repr__(self):
        return (
            "<ServerProxy for %s%s>" %
            (self.__host, self.__handler)
            )

    __str__ = __repr__

    def __getattr__(self, name):
        return dispatcher._Method(self._request, name, self._json_impl,
                                  self._send_id)




##############################################################################
#
# Copyright 2008 Lovely Systems GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################

from lovely.jsonrpc import wsgi
import threading
from wsgiref.simple_server import make_server
from lovely.jsonrpc import dispatcher
from lovely.jsonrpc import proxy

class TestingAPI(object):

    def echo(self, *args, **kwargs):
        return args, kwargs


class App(object):

    def __init__(self, wrapped):
        self.wrapped = wrapped

    def __call__(self, environ, start_response):
        for k in [k for k in sorted(environ) if k.startswith('HTTP_')]:
            print k, environ[k]
        res = self.wrapped(environ, lambda s, h: self.start_response(
            start_response, s, h))
        return res

    def start_response(self, sr, status, headers):
        headers += (('Set-Cookie', 'x=1'),)
        return sr(status, headers)

def get_server(port=12345):
    api = TestingAPI()
    app = wsgi.WSGIJSONRPCApplication({'':dispatcher.JSONRPCDispatcher(api)})
    return  make_server('localhost', 12345, App(app))

class OneRequest(threading.Thread):

    def __init__(self, server):
        super(OneRequest, self).__init__()
        self.server = server

    def run(self):
        self.server.handle_request()

def one_request(server):
    OneRequest(server).start()


class TestJSONRPCTransport(object):
    """transport for testing"""

    def __init__(self, base, app):
        self.base = base
        self._client = app

    def request(self, requestBody):
        self._client.lastRequestBody = requestBody
        resp = self._client.post(self.base, requestBody,
                                 content_type="application/json")
        return resp.status_int, resp.body

def TestJSONRPCProxy(base, app):
    import simplejson
    return proxy.ServerProxy(
        base,
        json_impl = simplejson,
        transport_impl=TestJSONRPCTransport,
        app=app)




##############################################################################
#
# Copyright 2008 Lovely Systems GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################

import unittest, doctest
from zope.testing.doctestunit import DocFileSuite, DocTestSuite

def test_suite():
    dispatcher = DocFileSuite('dispatcher.txt',
            optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
                             )
    proxy = DocFileSuite('proxy.txt',
            optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
                             )
    wsgi = DocFileSuite('wsgi.txt',
            optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
                             )
    s = unittest.TestSuite((dispatcher, proxy, wsgi))
    return s



##############################################################################
#
# Copyright 2008 Lovely Systems GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
##############################################################################

import types
import logging
import sys, traceback
import copy
import time

_log = logging.getLogger(__name__)

_descriptions = set(['summary', 'help', 'idempotent', 'params', 'return'])

class JSONRPCError(Exception):
    """JSONRPC Error"""

class BadJSONRequestError(Exception):
    """Cannot parse JSON-RPC request body"""

def describe_method(method):
    """Check is a callable object has description params set"""
    description = {}
    for key in _descriptions:
        if hasattr(method, key):
            description[key] = getattr(method, key)
    return description


class JSONRPCDispatcher(object):

    def __init__(self, instance=None, methods=None,
                 name='Python JSONRPC Service',
                 summary='Service dispatched by python JSONRPCDispatcher',
                 help=None, address=None, json_impl=None):

        """Initialization. Can take an instance to register upon initialization"""

        if json_impl is None:
            import simplejson
            self.json_impl = simplejson
        else:
            self.json_impl = json_impl
        self.instances = []
        self.name = name
        self.help = help
        self.address = address
        self.summary = summary
        # Store all attributes of class before any methods are added
        # for negative lookup later
        self.base_attributes = set(dir(self))
        self.base_attributes.add('base_attributes')

        # If instance was given during initialization then register it
        if instance is not None:
            self.register_instance(instance)
        if methods is not None:
            for method in methods:
                self.register_method(method)

        self.__dict__[u'system.list_methods'] = self.system_list_methods
        self.__dict__[u'system.describe'] = self.system_describe

    def get_valid_methods(self):
        valid_methods = {}
        for key, value in self.__dict__.items():
            if key.startswith('_') is False:
                if key not in self.base_attributes:
                    valid_methods[key] = value
        return valid_methods

    def register_instance(self, instance):
        """Registers all attributes of class instance to dispatcher"""
        for attribute in dir(instance):
            if attribute.startswith('_') is False:
                # All public attributes
                self.register_method(getattr(instance, attribute), name=attribute)

        # Store it in the list for convienience
        self.instances.append(instance)


    def register_method(self, function, name=None):
        """Registers a method with the dispatcher"""
        # If the name isn't passed try to find it. If we can't fail gracefully.
        if name is None:
            try:
                name = function.__name__
            except:
                if hasattr(function, '__call__'):
                    raise """Callable class instances must be passed with name parameter"""
                else:
                    raise """Not a function"""

        if self.__dict__.has_key(name) is False:
            self.__dict__[unicode(name)] = function
        else:
            print 'Attribute name conflict -- %s must be removed before attribute of the same name can be added'


    def system_list_methods(self):
        """List all the available methods and return a object parsable
        that conforms to the JSONRPC Service Procedure Description
        specification"""
        method_list = []
        for key, value in self.get_valid_methods().items():
            method = {}
            method['name'] = key
            method.update(describe_method(value))
            method_list.append(method)
        method_list.sort()
        _log.debug('system.list_methods created list %s' % str(method_list))
        return method_list

    def system_describe(self):
        """Service description"""
        description = {}
        description['sdversion'] = '1.0'
        description['name'] = self.name
        description['summary'] = self.summary
        if self.help is not None:
            description['help'] = self.help
        if self.address is not None:
            description['address'] = self.address
        description['procs'] = self.system_list_methods()
        return description

    def dispatch(self, json):
        """Public dispatcher, verifies that a method exists in it's
        method dictionary and calls it"""
        try:
            rpc_request = self._decode(json)
        except ValueError:
            raise BadJSONRequestError, json
        _log.debug('decoded to python object %s' % str(rpc_request))

        if self.__dict__.has_key(rpc_request[u'method']):
            _log.debug('dispatcher has key %s' % rpc_request[u'method'])
            return self._dispatch(rpc_request)
        else:
            _log.debug('returning jsonrpc error')
            return self._encode(result=None, error=JSONRPCError('no such method'))

    def _dispatch(self, rpc_request):
        """Internal dispatcher, handles all the error checking and
        calling of methods"""
        result = None
        error = None
        jsonrpc_id = None

        logged_failure = False

        params = rpc_request.get('params', [])
        args = []
        kwargs = {}
        if type(params) is types.DictType:
            sargs = []
            for k, v in params.items():
                k = str(k)
                try:
                    k = int(k)
                    sargs.append((k, v))
                except ValueError:
                    kwargs[k] = v
            args = [a[1] for a in sorted(sargs)]
        elif type(params) in (list, tuple):
            args =  params
        elif params is not None:
            # If type was something weird just return a JSONRPC Error
            raise JSONRPCError, 'params not array or object type'
        try:
            result = self.__dict__[rpc_request[u'method']](*args, **kwargs)
        except Exception, e:
            error = JSONRPCError('Server Exception :: %s' % e)
            error.type = e.__class__
            tb = ''.join(traceback.format_exception(*sys.exc_info()))
            logging.error(tb)

        if rpc_request.has_key('id'):
            jsonrpc_id = rpc_request[u'id']

        return self._encode(result=result, error=error, jsonrpc_id=jsonrpc_id)


    def _encode(self, result=None, error=None, jsonrpc_id=None):
        """Internal encoder method, handles error formatting, id
        persistence, and encoding via the give json_impl"""
        response = {}
        response['result'] = result
        if jsonrpc_id is not None:
            response['id'] = jsonrpc_id


        if error is not None:
            if hasattr(error, 'type'):
                error_type = str(error.type)
                error_message = str(error)
            else:
                error_type = 'JSONRPCError'
                error_message = str(error).strip()

            response['error'] = {'type':error_type,
                                 'message':error_message}
        _log.debug('serializing %s' % str(response))
        return self.json_impl.dumps(response)

    def _decode(self, json):
        """Internal method for decoding json objects, uses simplejson"""
        return self.json_impl.loads(json)

class _Method(object):

    def __init__(self, call, name, json_impl, send_id):
        self.call = call
        self.name = name
        self.json_impl = json_impl
        self.send_id = send_id

    def __call__(self, *args, **kwargs):
        # Need to handle keyword arguments per 1.1 spec
        request = {}
        request['version'] = '1.1'
        request['method'] = self.name
        if self.send_id:
            request['id'] = int(time.time())
        if len(kwargs) is not 0:
            params = copy.copy(kwargs)
            index = 0
            for arg in args:
                params[str(index)] = arg
                index = index + 1
        elif len(args) is not 0:
            params = copy.copy(args)
        else:
            params = None
        request['params'] = params
        _log.debug('Created python request object %s' % str(request))
        return self.call(self.json_impl.dumps(request))

    def __getattr__(self, name):
        return _Method(self.call, "%s.%s" % (self.name, name), self.json_impl)



import os
from setuptools import setup, find_packages

longDesc = """

This package includes json-rpc client and server (wsgi) implementation.

"""

setup(
    name="lovely.jsonrpc",
    version="0.1.0a1",
    license = "Apache License 2.0",
    description = 'Python JSONRPC package',
    long_description = longDesc,
    keywords = "JSON RPC JSONRPC WSGI",
    author = "Lovely Systems AG",
    author_email = "bernddorn@gmail.com",
    url = 'http://code.google.com/p/lovely-jsonrpc',
    packages=find_packages('src'),
    package_dir = {'':'src'},
    zip_safe = True,
    include_package_data = False,
    platforms = ('Any',),
    namespace_packages = ['lovely'],
    install_requires = ['setuptools'],
    extras_require = dict(test=['zope.testing',
                                'simplejson',
                                'webtest']),
    )
